#include "str8.h"
#include "aliases.h"
#include "mem_allocator.h"
#include <string.h>

Str8 *wapp_str8_buf_alloc(const Allocator *allocator, u64 capacity) {
  Str8 *str = NULL;

  if (!allocator) {
    goto RETURN_STR8;
  }

  str = wapp_mem_allocator_alloc(allocator, sizeof(Str8));
  if (!str) {
    goto RETURN_STR8;
  }

  c8 *buf = wapp_mem_allocator_alloc(allocator, sizeof(c8) * capacity);
  if (!buf) {
    wapp_mem_allocator_free(allocator, (void **)&str);
    goto RETURN_STR8;
  }

  str->buf      = buf;
  str->size     = 0;
  str->capacity = capacity;

RETURN_STR8:
  return str;
}

Str8 *wapp_str8_alloc_cstr(const Allocator *allocator, const char *str) {
  Str8 *output = NULL;

  if (!allocator || !str) {
    goto RETURN_ALLOC_CSTR;
  }

  u64 length = strlen(str);
  output     = wapp_str8_buf_alloc(allocator, length * 2);
  if (!output) {
    goto RETURN_ALLOC_CSTR;
  }

  output->size = length;
  memcpy(output->buf, str, length);

RETURN_ALLOC_CSTR:
  return output;
}

Str8 *wapp_str8_alloc_str8(const Allocator *allocator, Str8RO *str) {
  Str8 *output = NULL;

  if (!allocator || !str) {
    goto RETURN_ALLOC_STR8;
  }

  output = wapp_str8_buf_alloc(allocator, str->capacity);
  if (!output) {
    goto RETURN_ALLOC_STR8;
  }

  output->size = str->size;
  memcpy(output->buf, str->buf, str->size);

RETURN_ALLOC_STR8:
  return output;
}

c8 wapp_str8_get(const Str8 *str, u64 index) {
  if (index >= str->size) {
    return '\0';
  }

  return str->buf[index];
}

void wapp_str8_set(Str8 *str, u64 index, c8 c) {
  if (index >= str->size) {
    return;
  }

  str->buf[index] = c;
}

Str8RO wapp_str8_substr(Str8RO *str, u64 start, u64 end) {
  if (start >= str->size || start >= end) {
    return (Str8RO){0};
  }

  if (end > str->size) {
    end = str->size;
  }

  return (Str8RO){
    .capacity = end - start,
    .size     = end - start,
    .buf      = str->buf + start,
  };
}

i64 wapp_str8_find(Str8RO *str, Str8RO substr) {
  if (substr.size > str->size) {
    return -1;
  }

  for (u64 i = 0; i < str->size; ++i) {
    const c8 *sub = str->buf + i;
    if (memcmp(sub, substr.buf, substr.size) == 0) {
      return i;
    }
  }

  return -1;
}

i64 wapp_str8_rfind(Str8RO *str, Str8RO substr) {
  if (substr.size > str->size) {
    return -1;
  }

  for (i64 i = str->size - substr.size; i >= 0; --i) {
    const c8 *sub = str->buf + i;
    if (memcmp(sub, substr.buf, substr.size) == 0) {
      return i;
    }
  }

  return -1;
}
